/**
 * Created by godsong on 16/3/21.
 */
import ConfigManager from "./ConfigManager";
import StateMachine from "./StateMachine";
import Monitor from "./Monitor";

if (typeof window.RTCPeerConnection === 'undefined') {
    window.RTCPeerConnection = window.webkitRTCPeerConnection || window.mozRTCPeerConnection;
}
const _interval = 2000;
const _sampleSize = 15;
const _sampleInterval = 8000;
let _rTyp = /typ (\w+) /;
ConfigManager.setDefaultConfig('RTC', {

    optional: {
        "optional": [
            {"DtlsSrtpKeyAgreement": true},
            {"RtpDataChannels": false}
        ]
    }
});
export default class ConnectionManager extends StateMachine {
    constructor(parent) {
        super('peer', parent);
        this.candidateList = [];
        this.candidateTyp = {};
    }

    connect(turnServers, allowVideo) {
        let config = ConfigManager.get('RTC');
        let self = this;
        let constraints = {
            optional: [{
                'DtlsSrtpKeyAgreement': 'true'
            }]
            , mandatory: {
                OfferToReceiveAudio: true,
                OfferToReceiveVideo: allowVideo ? true : false
            }
        };
        this.peerConnection = new RTCPeerConnection({
            iceServers: turnServers || config.iceServers,
            iceTransportPolicy: 'relay'
        }, constraints);
        this.peerConnection.oniceconnectionstatechange = function (event) {
            self.nextState(event.target.iceConnectionState);
        };


        /* this.peer.onicecandidate=function(candidate){
         Monitor.log(candidate);
         }*/
        return this.peerConnection;
    }

    setRemoteSDP(sdp) {
        let self = this;
        if (Monitor.UA.ChromeVersion && Monitor.UA.ChromeVersion < 41) {
            sdp.sdp = sdp.sdp.replace(/audio \d UDP\/TLS\/RTP\/SAVPF/, "audio 1 RTP/SAVPF");
        }
        this.peerConnection.setRemoteDescription(new RTCSessionDescription(sdp), function () {
            Monitor.log('setRemoteSDP:');
            console.log('sdp:', sdp);
        }, function (e) {
            self.emitError('remoteSDPError', e);
        });
        //if (sdp.type == 'answer') {
        this.digestCandidate();
        //}
        if (this._remoteSDPReadyCallback) {
            this._remoteSDPReadyCallback();
            this._remoteSDPReadyCallback = null;
        }
    }

    digestCandidate() {
        let candidate;
        while (candidate = this.candidateList.shift()) {
            this._addIceCandidate(candidate);
        }

    }

    _addIceCandidate(candidate) {
        try {
            this.peerConnection.addIceCandidate(candidate);
        } catch (e) {
            this.emitError('candidateError');
        }
    }

    getRemoteStream() {
        let peer = this.peerConnection;
        return new Promise(function (resolve, reject) {
            peer.onaddstream = function (event) {
                Monitor.log('onaddstream');
                let mediaStream = event.stream;
                mediaStream.onended = function () {
                    Monitor.log('remote stream onend');
                };
                resolve(mediaStream);
            }
        })
    }

    addCandidate(candidateEmitter) {
        var peer = this.peerConnection;
        let self = this;
        candidateEmitter.on('value', function (candidate) {
            try {
                Monitor.log('add candidate');
                if (peer.remoteDescription && peer.remoteDescription.sdp) {
                    self._addIceCandidate(new RTCIceCandidate({
                        sdpMLineIndex: candidate.sdpMLineIndex,
                        candidate: candidate.candidate
                    }));
                }
                else {
                    self.candidateList.push(new RTCIceCandidate({
                        sdpMLineIndex: candidate.sdpMLineIndex,
                        candidate: candidate.candidate
                    }));
                }
            } catch (e) {
                console.error(e);
            }
        })
    }

    setSDP(SDPPromise) {
        let self = this;
        return SDPPromise.then(function (sdp) {
            self.setRemoteSDP(new RTCSessionDescription(sdp));
        })
    }

    createCandidate(sender) {
        let self = this;
        this.peerConnection.onicecandidate = function (event) {
            if (event.candidate) {
                let match = _rTyp.exec(event.candidate.candidate);
                if (match) {
                    if (isNaN(self.candidateTyp[match[1]])) {
                        self.candidateTyp[match[1]] = 1;
                    }
                    else {
                        self.candidateTyp[match[1]]++;
                    }
                }
                sender(event.candidate);
            }
            else if (event.candidate == null) {
                Monitor.log('candidate type:', self.candidateTyp);
            }
        }
    }

    createAnswer(sender) {
        var peer = this.peerConnection;
        let self = this;
        peer.createAnswer(function (offer) {
            Monitor.log('createAnswer');
            peer.setLocalDescription(offer, function () {
                Monitor.log('setLocalSDP');
            }, function (e) {
                self.emitError('remoteSDPError', e);
            });
            sender(offer);
            //self.digestCandidate();
        }, function (e) {
            self.emitError('answerError', e);
        })

    }

    remoteSDPReady() {
        //简易的promise 浏览器的promise风格不符合需求
        let self = this;
        return {
            then: function (callback) {
                if (self.peerConnection.remoteDescription && self.peerConnection.remoteDescription.sdp) {
                    callback();
                    return new Promise(1);
                }
                else {
                    return new Promise(function (resolve) {
                        self._remoteSDPReadyCallback = function () {
                            callback();
                            resolve();
                        };
                    })

                }
            }
        }
    }

    createOffer(sender) {
        let peer = this.peerConnection;
        let self = this;
        peer.createOffer(function (sdp) {
            Monitor.log('createOffer');
            peer.setLocalDescription(sdp, function () {
                Monitor.log('setLocalSDP');

            }, function (e) {
                self.emitError('localSDPError', e);
            });
            sender(sdp);

        }, function (e) {
            self.emitError('offerError', e);
        })
    }

    onConnected() {
        this.startMonitor();
    }

    close() {
        this.peerConnection && this.peerConnection.close();
        clearInterval(this._timer);
        this._timer = null;
        this.destroy();
    }

    startMonitor() {
        let self = this;
        let sample = [];
        let counter = 0;
        if (!this._timer) {
            let lastBytesSent = 0, lastBytesReceived = 0;
            this._timer = setInterval(function () {
                self.peerConnection.getStats(function (stats) {
                    let result = stats.result();
                    let report = {};
                    for (let i = 0; i < result.length; i++) {
                        let stat = result[i];
                        if (/ssrc_\d+_send/.test(stat.id) && stat.stat('googFrameHeightInput') !== '') {
                            report['frameHeightSent'] = stat.stat('googFrameHeightSent') || 0;
                            report['frameWidthSent'] = stat.stat('googFrameWidthSent') || 0;
                            report['frameRateSent'] = stat.stat('googFrameRateSent') || 0;
                            report['packetsSentLostRate'] = (stat.stat('packetsLost') / stat.stat('packetsSent') * 100).toFixed(2) + '%';
                            report['rtt'] = stat.stat('googRtt');
                            let sentByteRate = (stat.stat('bytesSent') - lastBytesSent) / (_interval / 1000) / 1024;
                            report['sentByteRate'] = sentByteRate;
                            report['sentByteRateText'] = _resolveBandwidthDisplay(sentByteRate);
                            lastBytesSent = stat.stat('bytesSent');

                        }
                        else if (/ssrc_\d+_recv/.test(stat.id) && stat.stat('googFrameHeightReceived') !== '') {
                            report['frameHeightReceived'] = stat.stat('googFrameHeightReceived') || 0;
                            report['frameWidthReceived'] = stat.stat('googFrameWidthReceived') || 0;
                            report['frameRateReceived'] = stat.stat('googFrameRateReceived') || 0;
                            report['delay'] = stat.stat('googCurrentDelayMs') || 0;
                            report['packetsReceivedLostRate'] = (stat.stat('packetsLost') / stat.stat('packetsReceived') * 100).toFixed(2) + '%';
                            let receivedByteRate = (stat.stat('bytesReceived') - lastBytesReceived) / (_interval / 1000) / 1024;
                            report['receivedByteRate'] = receivedByteRate;
                            report['receivedByteRateText'] = _resolveBandwidthDisplay(receivedByteRate);
                            lastBytesReceived = stat.stat('bytesReceived');
                        }
                        else if (/Conn-audio-\d+-\d+/.test(stat.id) && stat.stat('googActiveConnection') == 'true') {
                            report['localCandidateType'] = stat.stat('googLocalCandidateType');
                            report['remoteCandidateType'] = stat.stat('googRemoteCandidateType');
                            report['transportType'] = stat.stat('googTransportType');
                        }
                    }
                    sample.push(report);
                    counter += _interval;
                    if (sample.length > _sampleSize) {
                        sample.shift();
                    }
                    if (counter >= _sampleInterval) {
                        counter = 0;
                        let averageReport = _analyze(sample);
                        self.emit('averageStats', averageReport);
                    }

                    self.emit('stats', report);
                });
            }, _interval);
        }
    }
}
function _resolveBandwidthDisplay(value) {
    let text = '';
    if (value >= 1000) {
        text = (value / 1024).toFixed(2) + 'MB/s';
    }
    else {
        text = +value.toFixed(2) + 'KB/s';
    }
    value = value * 8;
    if (value >= 1000) {
        text += '(' + (value / 1024).toFixed(2) + 'Mbps)';
    }
    else {
        text += '(' + value.toFixed(2) + 'Kbps)';
    }
    return text;
}
function _analyze(sample) {
    let mapperSent = {},
        mapperRecv = {},
        maxSentFps = 0,
        maxRecvFps = 0,
        maxSentBR = 0,
        maxRecvBR = 0,
        maxSentLostPacket = 0,
        maxRecvLostPacket = 0;
    sample.forEach(function (report) {
        _analyzeResolution(mapperSent, report, 'Sent');
        _analyzeResolution(mapperRecv, report, 'Received');
        maxSentFps += +report['frameRateSent'];
        maxRecvFps += +report['frameRateReceived'];
        maxSentBR += +report['sentByteRate'];
        maxRecvBR += +report['receivedByteRate'];
        maxSentLostPacket += parseInt(report['packetsSentLostRate']);
        maxRecvLostPacket += parseInt(report['packetsReceivedLostRate']);
    });
    let sentResolution = _findAverageResolution(mapperSent);
    let recvResolution = _findAverageResolution(mapperRecv);
    return {
        frameWidthSent: +sentResolution[0],
        frameHeightSent: +sentResolution[1],
        frameWidthReceived: +recvResolution[0],
        frameHeightReceived: +recvResolution[1],
        frameRateSent: (maxSentFps / sample.length).toFixed(2),
        frameRateReceived: (maxRecvFps / sample.length).toFixed(2),
        receivedByteRate: Math.round(maxRecvBR / sample.length),
        receivedByteRateText: _resolveBandwidthDisplay(Math.round(maxRecvBR / sample.length)),
        sentByteRate: Math.round(maxSentBR / sample.length),
        sentByteRateText: _resolveBandwidthDisplay(Math.round(maxSentBR / sample.length)),
        packetsSentLostRate: (maxSentLostPacket / sample.length).toFixed(2) + '%',
        packetsReceivedLostRate: (maxRecvLostPacket / sample.length).toFixed(2) + '%'
    }
}
function _analyzeResolution(mapper, report, type) {
    let r = report['frameWidth' + type] + ',' + report['frameHeight' + type];
    if (!mapper[r]) {
        mapper[r] = 1;
    }
    else {
        mapper[r] += 1;
    }
}
function _findAverageResolution(mapper) {
    var list = [];
    for (var k in mapper) {
        if (mapper.hasOwnProperty(k)) {
            list.push({resolution: k.split(','), count: mapper[k]});
        }
    }
    list.sort(function (a, b) {
        return b.count - a.count;
    });
    return list[0].resolution;
}