/**
 * Created by godsong on 16/3/23.
 */
export default class EventEmitter {
    constructor() {
        this._eventHandlers = {};
        this._emitCaches = {};
    }

    on(method, handler) {
        if (!this._eventHandlers[method]) {
            this._eventHandlers[method] = [];
        }
        this._eventHandlers[method].push(handler);
        let self = this;
        setTimeout(function () {
            if (self._emitCaches[method] && self._emitCaches[method].length > 0) {
                let emitValue;

                while ((emitValue = self._emitCaches[method].shift()) && self._eventHandlers[method] && self._eventHandlers[method].length > 0) {
                    self.emit(method, emitValue);
                }
            }
        });
    }

    emit(method, value, notCache) {
        let self = this;
        let args = Array.prototype.slice.call(arguments, 1);
        if (this._eventHandlers[method] && this._eventHandlers[method].length > 0) {
            this._eventHandlers[method].forEach(function (handler) {
                handler.apply(self, args);
            });
            return true;
        }
        else {
            (this._emitCaches[method] || (this._emitCaches[method] = [])).push(value);
            return false;
        }
    }

    emitError(error, value) {
        if (!this.emit(error, value)) {
            this.emit('error', value, error);
        }
    }

    clearCache() {
        this._emitCaches = {};
    }

    off(method, handler) {
        if (arguments.length > 0) {
            if (handler) {
                for (let i = this._eventHandlers[method].length; i-- > 0;) {
                    if (this._eventHandlers[method][i] === handler) {
                        this._eventHandlers[method].splice(i, 1);
                        break;
                    }
                }
            }
            else {
                delete this._eventHandlers[method];
            }
        }
        else {
            this._eventHandlers = {};
        }
    }

    once(method, handler) {
        let self = this;
        let fired = false;

        function g() {
            this.off(method, g);
            if (!fired) {
                fired = true;
                handler.apply(self, Array.prototype.slice.call(arguments));
            }
        }

        this.on(method, g);
    }

}
