/**
 * Created by godsong on 16/3/21.
 */
import ConfigManager from "./ConfigManager";

if (typeof navigator.getUserMedia === 'undefined') {
    navigator.getUserMedia = navigator.webkitGetUserMedia || navigator.mozGetUserMedia;
}
if (typeof window.URL === 'undefined') {
    window.URL = window.webkitURL || window.mozURL;
}
//初始化media的默认配置
ConfigManager.setDefaultConfig('media', {
    constraints: {
        video: true,
        audio: true
    }
});
class UserMedia {
    constructor() {
        this.mediaList = [];
    }

    remove(media) {
        for (let i = 0; i < this.mediaList.length; i++) {
            if (this.mediaList[i] === media) {
                this.mediaList.splice(i, 1);
            }
        }
        if (this.mediaList.length == 0) {
            this.mediaStream = null;
        }
    }

    getUserMedia(media) {
        let self = this;
        this.mediaList.push(media);
        if (this.mediaStream) {
            _resolveMedia(this.mediaList, this.mediaStream);
        }
        if (!this.getting) {
            this.getting = true;
            navigator.getUserMedia(ConfigManager.get('media').constraints, function (mediaStream) {
                self.mediaStream = mediaStream;
                self.getting = false;
                _resolveMedia(self.mediaList, mediaStream);
            }, function (e) {
                self.getting = false;
                _rejectMedia(self.mediaList, e);
            });
        }
    }
}
export default class Media {
    constructor(videoElement) {
        if (typeof videoElement === 'string') {
            videoElement = document.getElementById(videoElement);
        }
        this.videoElement = videoElement;
        this.destinationList = [];
    }

    fetchMedia(remote) {
        let self = this;
        if (remote && remote.getRemoteStream) {
            remote.getRemoteStream().then(function (mediaStream) {
                self.onMedia(mediaStream);
            });

        }
        else {
            UserMediaManager.getUserMedia(this);
        }
    }

    onMedia(mediaStream) {
        this.mediaStream = mediaStream;
        if ('srcObject' in this.videoElement) {
            this.videoElement.srcObject = mediaStream;
        }

        else {
            this.videoElement.src = URL.createObjectURL(mediaStream);
            this.videoElement.setAttribute('autoplay', 'true');
        }
        _attach(this.destinationList, mediaStream);
        this.resolve && this.resolve(mediaStream);
    }

    onError(e) {
        this.error = e;
        this.reject && this.reject(e);
    }

    sendTo(peerConnection) {
        let self = this;
        this.destinationList.push({object: peerConnection, attched: false});
        return new Promise(function (resolve, reject) {
            if (self.mediaStream) {
                _attach(self.destinationList, self.mediaStream);
                resolve(self.mediaStream);
            }
            else if (self.error) {
                reject(self.error);
            }
            else {
                self.resolve = resolve;
                self.reject = reject;
            }
        })
    }

    setVideoMute(flag) {
        this.mediaStream.getVideoTracks().forEach(function (track) {
            track.enabled = !flag;
        });
    }

    toggleVideoMute() {
        let flag = true;
        this.mediaStream.getVideoTracks().forEach(function (track) {
            track.enabled = !track.enabled;
            if (track.enabled) {
                flag = false;
            }
        });
        return flag;
    }

    isVideoMute() {
        let flag = true;
        let tracks = this.mediaStream.getVideoTracks();
        for (let i = tracks.length; i-- > 0;) {
            if (tracks[i].enabled) {
                flag = false;
                break;
            }
        }
        return flag;
    }

    initMedia(allowVideo) {
        let mediaConfig = ConfigManager.get('media');
        mediaConfig.constraints.video = allowVideo;
    }

    setAudioMute(flag) {
        if (flag) {
            this.videoElement.originVolume = this.videoElement.volume;
            this.videoElement.volume = 0;
        }
        else {
            this.videoElement = this.videoElement.originVolume;
        }

    }

    close() {
        UserMediaManager.remove(this);
        this.destinationList.forEach(function (dest) {
            dest.oldStream && dest.object.removeStream(dest.oldStream);
        });
        this.videoElement.srcObj = null;
        this.videoElement.src = '';
        this.mediaStream && _destroyMediaStream(this.mediaStream);
    }
}

function _attach(destinationList, mediaStream) {
    destinationList.forEach(function (dest) {
        if (dest.attached) {
            dest.oldStream && dest.object.removeStream(dest.oldStream);
            dest.attached = false;
        }
        dest.object.addStream(mediaStream);
        dest.oldStream = mediaStream;
        dest.attached = true;
    })
}
const UserMediaManager = new UserMedia();

function _resolveMedia(mediaList, mediaStream) {
    mediaList.forEach(function (media) {
        media.onMedia(mediaStream);
    });
}
function _rejectMedia(mediaList, mediaStream, error) {
    mediaList.forEach(function (media) {
        media.onError(mediaStream, error);
    });
}


function _destroyMediaStream(mediaStream) {
    mediaStream.getTracks().forEach(function (track) {
        track.stop();
    })
}