/**
 * Created by godsong on 16/3/21.
 */
import ConfigManager from './ConfigManager';
import EventEmitter from './EventEmitter';
import SignalType from './SignalType';
import {CallStatus,CallType} from "./Enum";
import IPromise from "./Promise";
import Monitor from "./Monitor";
if (typeof WebSocket === 'undefined') {
    window.WebSocket = webkitWebSocket || mozWebSocket;
}
function _pruneLogData(data) {
    if (data !== undefined) {
        let pruneData = {};
        let thisUID = data.initiativeUserInfo && data.initiativeUserInfo.userId,
            thatUID = data.passtiveUserInfo ? data.passtiveUserInfo.userId : (data.userInfo && data.userInfo.userId || 0);
        if (thisUID)pruneData.thisUID = thisUID;
        if (thatUID) pruneData.thatUID = thatUID;
        if (data.roomId)pruneData.roomId = data.roomId;
        if (data.status != undefined)pruneData.status = data.status;
        if (data.sdpInfo)pruneData.sdpInfo = 1;
        if (data.type)pruneData.type = data.type;
        return pruneData;
    }

}
export default class Signaling extends EventEmitter {
    constructor() {
        super();
        let signaling = this;
        signaling._sessionInfo = null;
        signaling.candidateEmitter = new EventEmitter();
        signaling.reset();
        signaling.offlineCount=0;
        signaling.on('CS_PUSH_CANDIDATE', function (res) {

            if (!signaling._sessionInfo || res.userInfo.userId === signaling._sessionInfo.thatUserInfo.userId) {
                let candidate = {
                    candidate: res.candidateInfo.candidate,
                    sdpMid: res.candidateInfo.id,
                    sdpMLineIndex: res.candidateInfo.label,
                    userId: res.userInfo.userId
                };
                signaling.candidateEmitter.emit('value', candidate);
            }
        });
        let _rPrune=/>>.*?>>/g;
        signaling.on('CS_RES_ERROR_REQUEST_FORMAT', function (res) {
            res.errorMsg=res.errorMsg.replace(_rPrune,'');
            signaling.emitError('signalError', res);
        });
        signaling.on('CS_RES_ERROR_REQUEST_FORMAT', function (res) {
            res.errorMsg=res.errorMsg.replace(_rPrune,'');
            signaling.emitError('signalError', res);
        });
        signaling.on('CS_RES_ERROR_SYSTEM', function (res) {
            res.errorMsg=res.errorMsg.replace(_rPrune,'');
            signaling.emitError('signalError', res);
        });


    }

    connect() {
        let signaling = this;
        let promise = new IPromise(true);
        signaling._isSocketReady = false;
        this.clearCache();
        //fixme 本来并不想signaling跟session有任何耦合的 但是没办法 需要一些信息

        let myprotocal = location.protocol.toLocaleLowerCase().indexOf('https') > -1 ? 'wss' : 'ws';
        let url = ConfigManager.get('url');
        if(signaling.socket){
            signaling.socket.onopen=null;
            signaling.socket.onclose=null;
            signaling.socket.onerror=null;
            signaling.socket=null;
        }

        try {
            signaling.socket = new WebSocket(myprotocal + '://' + url);
        }catch(e){
            console.error(e);
        }
        if(signaling.socket) {
            signaling.socket.onmessage = function (e) {
                let obj = JSON.parse(e.data);
                if (obj.type) {
                    let method = SignalType.codeToMethod([obj.type]);
                    let data = obj[SignalType.getPackName(obj.type)];
                    if (method != 'CS_PUSH_CANDIDATE') {
                        Monitor.log('recv[' + method + ']->', _pruneLogData(data));
                    }
                    if (obj.type < 9000) {
                        signaling.emit(method, data);

                    } else {
                        signaling.emit(method, obj, obj.userId || null);

                    }
                }
                obj.msgId && signaling.send('ack', {ack: obj.msgId});
            };
            signaling.socket.onopen = function () {
                signaling.socketError = null;
                signaling._isSocketReady = true;
                clearInterval(signaling.heartBeatTimer);
                signaling.heartBeatTimer=setInterval(function(){
                    if(navigator.onLine==false&&signaling.socket.readyState==WebSocket.OPEN){
                        signaling.offlineCount++;
                    }
                    else{
                        signaling.offlineCount=0;
                    }
                    if(signaling.offlineCount>10){
                        signaling.offlineCount=0
                        signaling.socket.close();
                    }
                },2000);
                promise.resolve(signaling);
            };
            signaling.socket.onerror = function (e) {
                Monitor.log('socketError:');
                signaling.socketError = e;
                //if (!signaling._isSocketReady)reject(e);
            };

            signaling.socket.onclose = function (e) {
                Monitor.log('socket closed');
                signaling.emit('close');
                promise.reject(signaling.socketError);
                signaling.offlineCount=0;
                clearInterval(signaling.heartBeatTimer);
            };
        }
        return promise;
    }

    isSocketReady() {
        return this._isSocketReady;
    }
    waitIOSOnline(callback){
        this.once('CS_RES_CALL',function(res){
            if(res.status==CallStatus.IOS_ONLINE){
                callback();
            }
        })
    }
    send(method, params, userId) {
        if (method != 'ack') {
            Monitor.log('send[' + method + ']->', method != 'CS_REQ_CANDIDATE' ? _pruneLogData(params) : undefined);
        }
        let sigCode = SignalType.methodToCode(method);
        let sigPackName = SignalType.getPackName(sigCode);
        let _param = {};
        sigCode ? _param['type'] = sigCode : void(0);
        //如果method是ack，则当做ack消息处理，否则当做一般消息处理
        // m=='ack'?(_param=p):(_param[sigPackName] = p);
        // 发送消息带上userId
        if (method == 'ack') {
            _param = params
        } else {
            _param[sigPackName] = params;
            _param.userId = userId;
        }
        if (this.socket.readyState == WebSocket.CLOSED || this.socket.readyState == WebSocket.CLOSING) {
            this.emitError('socketError');
        }
        else {
            try {
                this.socket.send(JSON.stringify(_param));
            } catch (e) {
                this.emitError('socketError', e);
            }
        }
    }

    call(initiativeUserInfo, passtiveUserInfo, type, disableVideo) {
        let signaling = this;
        let promise = new IPromise();
        this._sessionInfo = {};
        _setSessionUserInfo(this._sessionInfo, initiativeUserInfo, passtiveUserInfo, true);
        signaling.send('CS_REQ_CALL', {
            initiativeUserInfo: initiativeUserInfo,
            passtiveUserInfo: passtiveUserInfo,
            allowVideo: disableVideo,
            type: type
        }, initiativeUserInfo.userId);
        function resCall(res) {
            if (res.status != CallStatus.OK)promise.resolve(res);

        }

        function pushRoom(res) {
            res=_clone(res);
            res.status = CallStatus.OK;
            res.turnServers = _resolveTurnServers(res.turnServers);
            signaling._sessionInfo.roomId = res.roomId;
            promise.resolve(res);
        }

        signaling.once('CS_RES_CALL', resCall);
        signaling.once('CS_PUSH_ROOM', pushRoom);
        return promise;
    }

    accept(sdp, roomId, initiativeUserInfo, passtiveUserInfo) {
        this.send('CS_RES_ACCEPT', {
            status: CallStatus.ACCEPT,
            initiativeUserInfo: initiativeUserInfo,
            passtiveUserInfo: passtiveUserInfo,
            allowVideo: true,
            sdpInfo: {type: 'answer', sdp: sdp},
            roomId: roomId
        }, passtiveUserInfo.userId)
    }

    reject(initiativeUserInfo, passtiveUserInfo, roomId, sdp) {
        this.send('CS_RES_ACCEPT', {
            status: CallStatus.REJECT,
            initiativeUserInfo: initiativeUserInfo,
            passtiveUserInfo: passtiveUserInfo,
            allowVideo: true,
            sdpInfo: {type: 'answer', sdp: sdp},
            roomId: roomId
        }, passtiveUserInfo.userId)
    }

    sayBusy(initiativeUserInfo, passtiveUserInfo, roomId) {
        this.send('CS_RES_ACCEPT', {
            status: CallStatus.ONLINE_BUSY,
            initiativeUserInfo: initiativeUserInfo,
            passtiveUserInfo: passtiveUserInfo,
            allowVideo: true,
            sdpInfo: {type: '', sdp: ''},
            roomId: roomId
        }, passtiveUserInfo.userId)
    }

    leaveRoom(roomId, userInfo, type) {
        this.send('CS_REQ_LEAVE_ROOM', {
            userInfo: userInfo,
            roomId: roomId,
            type: type
        }, userInfo.userId);
    }

    sendSDP(sdp, roomId, userInfo) {

        this.send('CS_REQ_SDP', {
            sdpInfo: {type: 'offer', sdp: sdp},
            userInfo: userInfo,
            roomId: roomId
        }, userInfo.userId);
    }

    getSDP() {

        let signaling = this;
        return new Promise(function (resolve, reject) {
            signaling.on('CS_PUSH_SDP', function (res) {
                if (!signaling._sessionInfo || res.userInfo.userId === signaling._sessionInfo.thatUserInfo.userId) {
                    resolve(res.sdpInfo);
                }
            })

        })
    }

    getCandidate() {
        return this.candidateEmitter;
    }

    closeCall(initiativeUserInfo, passtiveUserInfo) {
        this.send('CS_REQ_CALL', {
            initiativeUserInfo: initiativeUserInfo,
            passtiveUserInfo: passtiveUserInfo,
            allowVideo: true,
            type: CallType.CloseSession
        }, initiativeUserInfo.userId);
    }

    sendCandidate(candidate, roomId, userInfo) {
        this.send('CS_REQ_CANDIDATE', {
            candidateInfo: {
                candidate: candidate.candidate,
                id: candidate.sdpMid,
                label: candidate.sdpMLineIndex
            },
            roomId: roomId,
            userInfo: userInfo
        }, userInfo.userId);
    }

    waitAnswer(callback) {
        this.on('CS_RES_CALL', function (res) {
            if (res.status == CallStatus.OK || res.status == CallStatus.REJECT || res.status == CallStatus.ONLINE_BUSY || res.status == CallStatus.ONLINE) {
                if (res.status == CallStatus.OK && !res.sdpInfo) {
                    console.warn('res call error');
                    return;
                }
                callback(res);
            }

        });

    }

    close() {
        this.socket.close();
        this._eventHandlers = {};

    }

    waitCall(callback) {
        this.on('CS_PUSH_CALL', function (res) {
            res=_clone(res);
            res.isCloseCall = res.type == CallType.CloseSession;
            res.turnServers = _resolveTurnServers(res.turnServers);
            this._sessionInfo = {};
            _setSessionUserInfo(this._sessionInfo, res.initiativeUserInfo, res.passtiveUserInfo, false);
            this._sessionInfo.roomId = res.roomId;
            callback(res);
        });
        /* let emitter=new EventEmitter()
         this.on('CS_PUSH_CALL',function(res){
         emitter
         })*/
    }

    onLeave(callback) {
        let signaling = this;
        signaling.on('CS_BROADCAST_LEAVE_ROOM', function (response) {
            if (!signaling._sessionInfo || signaling._sessionInfo.roomId == response.roomId) {
                callback(response.type);
            }
        });
    }
    returnSDPTime(){
        let signaling=this;
        this.on('CS_PUSH_DATA', function (response) {
                if(response.data==='returnSDPtime'){
                   this.send('CS_REQ_DATA',{
                       roomId:signaling._sessionInfo.roomId,
                       data:'returnSDPtime',
                       userInfo:signaling._sessionInfo.thisUserInfo
                   })
                }
        });
    }
    onIPInfo(callback){
        this.on('CS_PUSH_DATA', function (response) {
            try{
                let data=JSON.parse(response.data);
                if(data.rawDataType==='IpInfo'){
                    callback(data);
                }
            }
            catch(e){

            }
        });
    }
    sendIpInfo(initiativeUserInfo,passtiveUserInfo){
        let data={
            initiativeUserIpInfo:initiativeUserInfo.ipInfo,
            initiativeUserIp:initiativeUserInfo.ip,
            passiveUserIpInfo:passtiveUserInfo.ipInfo,
            passiveUserIp:passtiveUserInfo.ip,
            rawDataType:'IpInfo'
        };
        this.send('CS_REQ_DATA',{
            roomId:this._sessionInfo.roomId,
            data:JSON.stringify(data),
            userInfo:this._sessionInfo.thisUserInfo
        })
    }
    onIceRestart(callback) {
        this.on('CS_PUSH_DATA', function (response) {
            let str = "ICE_RESTART:";
            if (response.data.indexOf(str) != -1) {
                let sdpString = response.data.substring(str.length);
                callback(sdpString);
            }

        });
    }

    reset() {
        this.off('CS_PUSH_CALL');
        this.off('CS_RES_CALL');
        this.off('CS_PUSH_ROOM');
        this.off('CS_PUSH_SDP');
        this.off('CS_BROADCAST_LEAVE_ROOM');
        this.off('CS_PUSH_DATA');
        this.candidateEmitter.off();
        this._sessionInfo = null;
        this.offlineCount=0;

    }

}
function _resolveTurnServers(tsv) {

    if (tsv) {
        let turnServers = [];
        for (let i = 0; i < tsv.uris.length; i++) {
            turnServers[i] = {
                username: tsv.username,
                url: tsv.uris[i],
                credential: tsv.password
            }
        }
        return turnServers;
    }

}
function _setSessionUserInfo(target, initiativeUserInfo, passtiveUserInfo, isCaller) {
    target.initiativeUserInfo = initiativeUserInfo;
    target.passtiveUserInfo = passtiveUserInfo;
    if (isCaller) {
        target.thisUserInfo = initiativeUserInfo;
        target.thatUserInfo = passtiveUserInfo;
    }
    else {
        target.thisUserInfo = passtiveUserInfo;
        target.thatUserInfo = initiativeUserInfo;
    }
}
function _clone(obj){
    let target={};
    for(let key in obj){
        if(obj.hasOwnProperty(key)){
            target[key]=obj[key];
        }
    }
    return target;
}